from datetime import timedelta

from django.test import TestCase
from django.utils import timezone
from dynamic_preferences.registries import global_preferences_registry

from .models import Helper, Message, Room, Shift, ShiftRegistration


class ShiftOverlapTests(TestCase):
    def setUp(self):
        # Create a room
        self.room = Room.objects.create(
            name="Test Room", required_helpers=1, meeting_location="Test Location"
        )

        # Create a helper
        self.helper = Helper.objects.create(
            phone="+491234567890", name="Test Helper", number_validated=True
        )

        # Create a base shift for testing overlaps
        self.base_shift = Shift.objects.create(
            room=self.room,
            start_at=timezone.now() + timedelta(hours=1),
            duration=timedelta(hours=2),
            required_helpers=1,
        )

        # Register the helper for the base shift
        self.base_registration = ShiftRegistration.objects.create(
            shift=self.base_shift,
            helper=self.helper,
            state=ShiftRegistration.RegState.REGISTERED,
        )

    def test_no_overlap_before(self):
        """Test a shift that ends before the base shift starts"""
        shift_before = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at - timedelta(hours=3),
            duration=timedelta(hours=1),
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_before))

    def test_back_to_back_shifts(self):
        """Test that a shift starting exactly when another ends is allowed"""
        shift_after = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at + self.base_shift.duration,
            duration=timedelta(hours=1),
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_after))

        # Also test the reverse case - registering for a shift that ends exactly when another begins
        self.base_registration.delete()
        ShiftRegistration.objects.create(
            shift=shift_after,
            helper=self.helper,
            state=ShiftRegistration.RegState.REGISTERED,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(self.base_shift))

    def test_no_overlap_after(self):
        """Test a shift that starts after the base shift ends"""
        shift_after = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at
            + self.base_shift.duration
            + timedelta(hours=1),
            duration=timedelta(hours=1),
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_after))

    def test_overlap_start(self):
        """Test a shift that starts during the base shift"""
        shift_overlap_start = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at + timedelta(hours=1),
            duration=timedelta(hours=2),
            required_helpers=1,
        )
        self.assertIsNotNone(self.helper.has_overlapping_shift(shift_overlap_start))

    def test_overlap_end(self):
        """Test a shift that ends during the base shift"""
        shift_overlap_end = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at - timedelta(hours=1),
            duration=timedelta(hours=2),
            required_helpers=1,
        )
        self.assertIsNotNone(self.helper.has_overlapping_shift(shift_overlap_end))

    def test_overlap_contained(self):
        """Test a shift that is completely contained within the base shift"""
        shift_contained = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at + timedelta(minutes=30),
            duration=timedelta(hours=1),
            required_helpers=1,
        )
        self.assertIsNotNone(self.helper.has_overlapping_shift(shift_contained))

    def test_overlap_contains(self):
        """Test a shift that completely contains the base shift"""
        shift_contains = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at - timedelta(hours=1),
            duration=timedelta(hours=4),
            required_helpers=1,
        )
        self.assertIsNotNone(self.helper.has_overlapping_shift(shift_contains))

    def test_exact_same_time(self):
        """Test a shift that has exactly the same time as the base shift"""
        shift_same_time = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at,
            duration=self.base_shift.duration,
            required_helpers=1,
        )
        self.assertIsNotNone(self.helper.has_overlapping_shift(shift_same_time))

    def test_deleted_shift_no_overlap(self):
        """Test that deleted shifts are not considered for overlap"""
        self.base_shift.deleted = True
        self.base_shift.save()

        shift_same_time = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at,
            duration=self.base_shift.duration,
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_same_time))

    def test_cancelled_registration_no_overlap(self):
        """Test that cancelled registrations are not considered for overlap"""
        self.base_registration.state = ShiftRegistration.RegState.CANCELED
        self.base_registration.save()

        shift_same_time = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at,
            duration=self.base_shift.duration,
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_same_time))

    def test_failed_registration_no_overlap(self):
        """Test that failed registrations are not considered for overlap"""
        self.base_registration.state = ShiftRegistration.RegState.FAILED
        self.base_registration.save()

        shift_same_time = Shift.objects.create(
            room=self.room,
            start_at=self.base_shift.start_at,
            duration=self.base_shift.duration,
            required_helpers=1,
        )
        self.assertIsNone(self.helper.has_overlapping_shift(shift_same_time))


class ShiftSignalTests(TestCase):
    def setUp(self):
        # Create rooms
        self.room1 = Room.objects.create(
            name="Test Room 1", required_helpers=1, meeting_location="Test Location 1"
        )
        self.room2 = Room.objects.create(
            name="Test Room 2", required_helpers=1, meeting_location="Test Location 2"
        )

        # Create a helper
        self.helper = Helper.objects.create(
            phone="+491234567890", name="Test Helper", number_validated=True
        )

        # Create a shift
        self.shift = Shift.objects.create(
            room=self.room1,
            start_at=timezone.now() + timedelta(hours=1),
            duration=timedelta(hours=2),
            required_helpers=1,
        )

        # Register the helper for the shift
        self.registration = ShiftRegistration.objects.create(
            shift=self.shift,
            helper=self.helper,
            state=ShiftRegistration.RegState.REGISTERED,
        )

        # Get global preferences manager
        self.global_preferences = global_preferences_registry.manager()

    def test_room_change_notification_default(self):
        """Test that room changes create notifications by default"""
        # Ensure the preference is False (default)
        self.global_preferences["helper__supress_room_change_notifications"] = False

        # Change the room
        self.shift.room = self.room2
        self.shift.save()

        # Check that a notification was created
        self.assertEqual(Message.objects.count(), 1)
        message = Message.objects.first()
        self.assertEqual(message.to, self.helper)
        self.assertTrue("geändert" in message.text)

    def test_room_change_notification_suppressed(self):
        """Test that room changes don't create notifications when suppressed"""
        # Enable notification suppression
        self.global_preferences["helper__supress_room_change_notifications"] = True

        # Change the room
        self.shift.room = self.room2
        self.shift.save()

        # Check that no notification was created
        self.assertEqual(Message.objects.count(), 0)

    def test_other_changes_still_notify_when_suppressed(self):
        """Test that other changes still create notifications when room notifications are suppressed"""
        # Enable notification suppression
        self.global_preferences["helper__supress_room_change_notifications"] = True

        # Change something other than the room
        self.shift.start_at = timezone.now() + timedelta(hours=2)
        self.shift.save()

        # Check that a notification was created
        self.assertEqual(Message.objects.count(), 1)
        message = Message.objects.first()
        self.assertEqual(message.to, self.helper)
        self.assertTrue("geändert" in message.text)
